//*****************************************************************************
// FILE:            WinMTRNet.cpp
//
//*****************************************************************************
#include "WinMTRGlobal.h"
#include "WinMTRNet.h"
#include "WinMTRDialog.h"
#include <iostream>
#include <sstream>
#include <string>

#include "stdio.h"
#include "string.h"

#define QQWRY "QQWry.dat"
#define REDIRECT_MODE_1 0x01
#define REDIRECT_MODE_2 0x02
#define MAXBUF 255


#define TRACE_MSG(msg)										\
	{														\
	std::ostringstream dbg_msg(std::ostringstream::out);	\
	dbg_msg << msg << std::endl;							\
	OutputDebugString(dbg_msg.str().c_str());				\
	}

#define IPFLAG_DONT_FRAGMENT	0x02
#define MAX_HOPS				30

struct trace_thread {
	int			address;
	WinMTRNet	*winmtr;
	int			ttl;
};

struct dns_resolver_thread {
	int			index;
	WinMTRNet	*winmtr;
};

void TraceThread(void *p);
void DnsResolverThread(void *p);

WinMTRNet::WinMTRNet(WinMTRDialog *wp) {
	
	ghMutex = CreateMutex(NULL, FALSE, NULL);
	tracing=false;
	initialized = false;
	wmtrdlg = wp;
	WSADATA wsaData;

    if( WSAStartup(MAKEWORD(2, 2), &wsaData) ) {
        AfxMessageBox("Failed initializing windows sockets library!");
		return;
    }

    hICMP_DLL =  LoadLibrary(_T("ICMP.DLL"));
    if (hICMP_DLL == 0) {
        AfxMessageBox("Failed: Unable to locate ICMP.DLL!");
        return;
    }

    /* 
     * Get pointers to ICMP.DLL functions
     */
    lpfnIcmpCreateFile  = (LPFNICMPCREATEFILE)GetProcAddress(hICMP_DLL,"IcmpCreateFile");
    lpfnIcmpCloseHandle = (LPFNICMPCLOSEHANDLE)GetProcAddress(hICMP_DLL,"IcmpCloseHandle");
    lpfnIcmpSendEcho    = (LPFNICMPSENDECHO)GetProcAddress(hICMP_DLL,"IcmpSendEcho");
    if ((!lpfnIcmpCreateFile) || (!lpfnIcmpCloseHandle) || (!lpfnIcmpSendEcho)) {
        AfxMessageBox("Wrong ICMP.DLL system library !");
        return;
    }

    /*
     * IcmpCreateFile() - Open the ping service
     */
    hICMP = (HANDLE) lpfnIcmpCreateFile();
    if (hICMP == INVALID_HANDLE_VALUE) {
        AfxMessageBox("Error in ICMP.DLL !");
        return;
    }

	ResetHops();

	initialized = true;
	return;
}

WinMTRNet::~WinMTRNet()
{
	if(initialized) {
		/*
		 * IcmpCloseHandle - Close the ICMP handle
		 */
		lpfnIcmpCloseHandle(hICMP);

		// Shut down...
		FreeLibrary(hICMP_DLL);

		WSACleanup();
	
		CloseHandle(ghMutex);
	}
}

void WinMTRNet::ResetHops()
{
	for(int i = 0; i < MaxHost;i++) {
		host[i].addr = 0;
		host[i].xmit = 0;
		host[i].returned = 0;
		host[i].total = 0;
		host[i].last = 0;
		host[i].best = 0;
		host[i].worst = 0;
		memset(host[i].name,0,sizeof(host[i].name));
	}
}

void WinMTRNet::DoTrace(int address)
{
	HANDLE hThreads[MAX_HOPS];
	tracing = true;

	ResetHops();

	last_remote_addr = address;

	// one thread per TTL value
	for(int i = 0; i < MAX_HOPS; i++) {
		trace_thread *current = new trace_thread;
		current->address = address;
		current->winmtr = this;
		current->ttl = i + 1;
		hThreads[i] = (HANDLE)_beginthread(TraceThread, 0 , current);
	}

	WaitForMultipleObjects(MAX_HOPS, hThreads, TRUE, INFINITE);
}

void WinMTRNet::StopTrace()
{
	tracing = false;
}

void TraceThread(void *p)
{
	trace_thread* current = (trace_thread*)p;
	WinMTRNet *wmtrnet = current->winmtr;
	TRACE_MSG("Threaad with TTL=" << current->ttl << " started.");

    IPINFO			stIPInfo, *lpstIPInfo;
    DWORD			dwReplyCount;
	char			achReqData[8192];
	int				nDataLen									= wmtrnet->wmtrdlg->pingsize;
	char			achRepData[sizeof(ICMPECHO) + 8192];


    /*
     * Init IPInfo structure
     */
    lpstIPInfo				= &stIPInfo;
    stIPInfo.Ttl			= current->ttl;
    stIPInfo.Tos			= 0;
    stIPInfo.Flags			= IPFLAG_DONT_FRAGMENT;
    stIPInfo.OptionsSize	= 0;
    stIPInfo.OptionsData	= NULL;

    for (int i=0; i<nDataLen; i++) achReqData[i] = 32; //whitespaces

    while(wmtrnet->tracing) {
	    
		// For some strange reason, ICMP API is not filling the TTL for icmp echo reply
		// Check if the current thread should be closed
		if( current->ttl > wmtrnet->GetMax() ) break;

		// NOTE: some servers does not respond back everytime, if TTL expires in transit; e.g. :
		// ping -n 20 -w 5000 -l 64 -i 7 www.chinapost.com.tw  -> less that half of the replies are coming back from 219.80.240.93
		// but if we are pinging ping -n 20 -w 5000 -l 64 219.80.240.93  we have 0% loss
		// A resolution would be:
		// - as soon as we get a hop, we start pinging directly that hop, with a greater TTL
		// - a drawback would be that, some servers are configured to reply for TTL transit expire, but not to ping requests, so,
		// for these servers we'll have 100% loss
		dwReplyCount = wmtrnet->lpfnIcmpSendEcho(wmtrnet->hICMP, current->address, achReqData, nDataLen, lpstIPInfo, achRepData, sizeof(achRepData), ECHO_REPLY_TIMEOUT);

		PICMPECHO icmp_echo_reply = (PICMPECHO)achRepData;

		wmtrnet->AddXmit(current->ttl - 1);
		if (dwReplyCount != 0) {
			TRACE_MSG("TTL " << current->ttl << " reply TTL " << icmp_echo_reply->Options.Ttl << " Status " << icmp_echo_reply->Status << " Reply count " << dwReplyCount);

			switch(icmp_echo_reply->Status) {
				case IP_SUCCESS:
				case IP_TTL_EXPIRED_TRANSIT:
					wmtrnet->SetLast(current->ttl - 1, icmp_echo_reply->RoundTripTime);
					wmtrnet->SetBest(current->ttl - 1, icmp_echo_reply->RoundTripTime);
					wmtrnet->SetWorst(current->ttl - 1, icmp_echo_reply->RoundTripTime);
					wmtrnet->AddReturned(current->ttl - 1);
					wmtrnet->SetAddr(current->ttl - 1, icmp_echo_reply->Address);
				break;
				case IP_BUF_TOO_SMALL:
					wmtrnet->SetName(current->ttl - 1, "Reply buffer too small.");
				break;
				case IP_DEST_NET_UNREACHABLE:
					wmtrnet->SetName(current->ttl - 1, "Destination network unreachable.");
				break;
				case IP_DEST_HOST_UNREACHABLE:
					wmtrnet->SetName(current->ttl - 1, "Destination host unreachable.");
				break;
				case IP_DEST_PROT_UNREACHABLE:
					wmtrnet->SetName(current->ttl - 1, "Destination protocol unreachable.");
				break;
				case IP_DEST_PORT_UNREACHABLE:
					wmtrnet->SetName(current->ttl - 1, "Destination port unreachable.");
				break;
				case IP_NO_RESOURCES:
					wmtrnet->SetName(current->ttl - 1, "Insufficient IP resources were available.");
				break;
				case IP_BAD_OPTION:
					wmtrnet->SetName(current->ttl - 1, "Bad IP option was specified.");
				break;
				case IP_HW_ERROR:
					wmtrnet->SetName(current->ttl - 1, "Hardware error occurred.");
				break;
				case IP_PACKET_TOO_BIG:
					wmtrnet->SetName(current->ttl - 1, "Packet was too big.");
				break;
				case IP_REQ_TIMED_OUT:
					wmtrnet->SetName(current->ttl - 1, "Request timed out.");
				break;
				case IP_BAD_REQ:
					wmtrnet->SetName(current->ttl - 1, "Bad request.");
				break;
				case IP_BAD_ROUTE:
					wmtrnet->SetName(current->ttl - 1, "Bad route.");
				break;
				case IP_TTL_EXPIRED_REASSEM:
					wmtrnet->SetName(current->ttl - 1, "The time to live expired during fragment reassembly.");
				break;
				case IP_PARAM_PROBLEM:
					wmtrnet->SetName(current->ttl - 1, "Parameter problem.");
				break;
				case IP_SOURCE_QUENCH:
					wmtrnet->SetName(current->ttl - 1, "Datagrams are arriving too fast to be processed and datagrams may have been discarded.");
				break;
				case IP_OPTION_TOO_BIG:
					wmtrnet->SetName(current->ttl - 1, "An IP option was too big.");
				break;
				case IP_BAD_DESTINATION:
					wmtrnet->SetName(current->ttl - 1, "Bad destination.");
				break;
				case IP_GENERAL_FAILURE:
					wmtrnet->SetName(current->ttl - 1, "General failure.");
				break;
				default:
					wmtrnet->SetName(current->ttl - 1, "General failure.");
			}

			if(wmtrnet->wmtrdlg->interval * 1000 > icmp_echo_reply->RoundTripTime)
				Sleep(wmtrnet->wmtrdlg->interval * 1000 - icmp_echo_reply->RoundTripTime);
		}

    } /* end ping loop */

	TRACE_MSG("Thread with TTL=" << current->ttl << " stopped.");

	delete p;
	_endthread();
}

int WinMTRNet::GetAddr(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int addr = ntohl(host[at].addr);
	ReleaseMutex(ghMutex);
	return addr;
}

int WinMTRNet::GetName(int at, char *n)
{
	WaitForSingleObject(ghMutex, INFINITE);
	if(!strcmp(host[at].name, "")) {
		int addr = GetAddr(at);
		sprintf (	n, "%d.%d.%d.%d", 
							(addr >> 24) & 0xff, 
							(addr >> 16) & 0xff, 
							(addr >> 8) & 0xff, 
							addr & 0xff
		);
		if(addr==0)
			strcpy(n,"");
	} else {
		strcpy(n, host[at].name);
	}
	ReleaseMutex(ghMutex);
	return 0;
}

int WinMTRNet::GetBest(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].best;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetWorst(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].worst;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetAvg(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].returned == 0 ? 0 : host[at].total / host[at].returned;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetPercent(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = (host[at].xmit == 0) ? 0 : (100 - (100 * host[at].returned / host[at].xmit));
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetLast(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].last;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetReturned(int at)
{ 
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].returned;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetXmit(int at)
{ 
	WaitForSingleObject(ghMutex, INFINITE);
	int ret = host[at].xmit;
	ReleaseMutex(ghMutex);
	return ret;
}

int WinMTRNet::GetMax()
{
	WaitForSingleObject(ghMutex, INFINITE);
	int max = MAX_HOPS;

	// first match: traced address responds on ping requests, and the address is in the hosts list
	for(int i = 0; i < MAX_HOPS; i++) {
		if(host[i].addr == last_remote_addr) {
			max = i + 1;
			break;
		}
	}

	// second match:  traced address doesn't responds on ping requests
	if(max == MAX_HOPS) {
		while((max > 1) && (host[max - 1].addr == host[max - 2].addr) && (host[max - 1].addr != 0) ) max--;
	}

	ReleaseMutex(ghMutex);
	return max;
}

void WinMTRNet::SetAddr(int at, __int32 addr)
{
	WaitForSingleObject(ghMutex, INFINITE);
	if(host[at].addr == 0 && addr != 0) {
		TRACE_MSG("Start DnsResolverThread for new address " << addr << ". Old addr value was " << host[at].addr);
		host[at].addr = addr;
		dns_resolver_thread *dnt = new dns_resolver_thread;
		dnt->index = at;
		dnt->winmtr = this;
		if(wmtrdlg->useDNS) _beginthread(DnsResolverThread, 0, dnt);
	}

	ReleaseMutex(ghMutex);
}

void WinMTRNet::SetName(int at, char *n)
{
	WaitForSingleObject(ghMutex, INFINITE);
	strcpy(host[at].name, n);
	ReleaseMutex(ghMutex);
}

void WinMTRNet::SetBest(int at, int current)
{
	WaitForSingleObject(ghMutex, INFINITE);
	if(host[at].best > current || host[at].xmit == 1) {
		host[at].best = current;
	};
	if(host[at].worst < current) {
		host[at].worst = current;
	}

	ReleaseMutex(ghMutex);
}

void WinMTRNet::SetWorst(int at, int current)
{
	WaitForSingleObject(ghMutex, INFINITE);
	ReleaseMutex(ghMutex);
}

void WinMTRNet::SetLast(int at, int last)
{
	WaitForSingleObject(ghMutex, INFINITE);
	host[at].last = last;
	host[at].total += last;
	ReleaseMutex(ghMutex);
}

void WinMTRNet::AddReturned(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	host[at].returned++;
	ReleaseMutex(ghMutex);
}

void WinMTRNet::AddXmit(int at)
{
	WaitForSingleObject(ghMutex, INFINITE);
	host[at].xmit++;
	ReleaseMutex(ghMutex);
}

/*unsigned long getValue( ȡļָ16ƴֵ
FILE *fp, ָļָ
unsigned long start, ָļƫ
int length) ȡ16ַ/
*/
unsigned long getValue(FILE *fp, unsigned long start, int length)
{
    unsigned long variable = 0;
    long val[255], i;

    fseek(fp, start, SEEK_SET);
    for (i = 0; i<length; i++)
    {
        /*˸λһζȡһַ*/
        val[i] = fgetc(fp) & 0x000000FF;
    }
    for (i = length - 1; i >= 0; i--)
    {
        /*Ϊȡ16ַ*/
        variable = variable * 0x100 + val[i];
    }
    return variable;
};


/*int getString( ȡļַַָ
FILE *fp, ָļָ
unsigned long start, ָļƫ
char **string) Žȡַַռ׵ַ
*/
int getString(FILE *fp, unsigned long start, char **string)
{
    unsigned long i = 0;
    char val;
    fseek(fp, start, SEEK_SET);
    /*ȡֱַ0x00Ϊֹ*/
    do
    {
        val = fgetc(fp);
        /*η洢ַռ*/
        *(*string + i) = val;
        i++;
    } while (val != 0x00);
    /*ַ*/
    return i;
};


/*void getAddress( ȡָIPĹλú͵λ
FILE *fp, ָļָ
unsigned long start, ָIPеļƫ
char **country, Źλõַռ׵ַ
char **location) ŵλõַռ׵ַ
*/
void getAddress(FILE *fp, unsigned long start, char **country, char **location)
{
    unsigned long redirect_address, counrty_address, location_address;
    char val;

    start += 4;
    fseek(fp, start, SEEK_SET);
    /*ȡ׵ֵַ*/
    val = (fgetc(fp) & 0x000000FF);

    if (val == REDIRECT_MODE_1)
    {
        /*ض1͵*/
        redirect_address = getValue(fp, start + 1, 3);
        fseek(fp, redirect_address, SEEK_SET);
        /*ͣض1ͽض2
        ȡضݣõλõļƫ*/
        if ((fgetc(fp) & 0x000000FF) == REDIRECT_MODE_2)
        {
            counrty_address = getValue(fp, redirect_address + 1, 3);
            location_address = redirect_address + 4;
            getString(fp, counrty_address, country);
        }
        /*ȡض1ݣõλõļƫ*/
        else
        {
            counrty_address = redirect_address;
            location_address = redirect_address + getString(fp, counrty_address, country);
        }
    }
    /*ض2͵*/
    else if (val == REDIRECT_MODE_2)
    {
        counrty_address = getValue(fp, start + 1, 3);
        location_address = start + 4;
        getString(fp, counrty_address, country);
    }
    else
    {
        counrty_address = start;
        location_address = counrty_address + getString(fp, counrty_address, country);
    }

    /*ȡλ*/
    fseek(fp, location_address, SEEK_SET);
    if ((fgetc(fp) & 0x000000FF) == REDIRECT_MODE_2 || (fgetc(fp) & 0x000000FF) == REDIRECT_MODE_1)
    {
        location_address = getValue(fp, location_address + 1, 3);
    }
    getString(fp, location_address, location);

    return;
};


/*void getHead( ȡֵķΧļͷУȵ28λ16ƣ
FILE *fp, ָļָ
unsigned long *start, ļƫֹλ
unsigned long *end) ļƫĽλ
*/
void getHead(FILE *fp, unsigned long *start, unsigned long *end)
{
    /*ֹλõļƫ洢ļͷеǰ816
    ƫΪ0ȡ4ַ*/
    *start = getValue(fp, 0L, 4);
    /*Ľλõļƫ洢ļͷеĵ81516
    ƫΪ4ַٶȡ4ַ*/
    *end = getValue(fp, 4L, 4);
};


/*unsigned long searchIP( ָIPλãöֲҷ
IPļƫ
һ¼Ľǣǰ416ƱʾʼIPַ
316ƣʾʼIPIPϢеλãļƫ
FILE *fp,
unsigned long index_start, ʼλõļƫ
unsigned long index_end, λõļƫ
unsigned long ip) ؼ֣ҪIP
*/
unsigned long searchIP(FILE *fp, unsigned long index_start, \

    unsigned long index_end, unsigned long ip)
{
    unsigned long index_current, index_top, index_bottom;
    unsigned long record;
    index_bottom = index_start;
    index_top = index_end;
    /*˴7Ϊһ¼ĳ7*/
    index_current = ((index_top - index_bottom) / 7 / 2) * 7 + index_bottom;
    /*ֲҷ*/
    do{
        record = getValue(fp, index_current, 4);
        if (record>ip)
        {
            index_top = index_current;
            index_current = ((index_top - index_bottom) / 14) * 7 + index_bottom;
        }
        else
        {
            index_bottom = index_current;
            index_current = ((index_top - index_bottom) / 14) * 7 + index_bottom;
        }
    } while (index_bottom<index_current);
    /*عؼIPļƫ*/
    return index_current;
};


/*unsigned long putAll( IPϢļļУص
FILE *fp,
FILE *out, ļָ룬ӵдȨ
unsigned long index_start, ʼļƫ
unsigned long index_end) Ľļƫ
*/
unsigned long putAll(FILE *fp, FILE *out, unsigned long index_start, unsigned long index_end)
{
    unsigned long i, count = 0;
    unsigned long start_ip, end_ip;
    char *country;
    char *location;

    country = (char*)malloc(255);
    location = (char*)malloc(255);

    /*˴7Ϊһ¼ĳ7*/
    for (i = index_start; i<index_end; i += 7)
    {
        /*ȡIPεʼIPͽIP
        ʼIPΪֵǰ4λ16
        IPIPϢֵǰ4λ16УָƫѰ*/
        start_ip = getValue(fp, i, 4);
        end_ip = getValue(fp, getValue(fp, i + 4, 3), 4);
        /*IPϢʽ
        ʼIP\tIP\tλ\tλ\n*/
        fprintf(out, "%d.%d.%d.%d", (start_ip & 0xFF000000) >> 0x18, \

            (start_ip & 0x00FF0000) >> 0x10, (start_ip & 0x0000FF00) >> 0x8, start_ip & 0x000000FF);
        fprintf(out, "\t");
        fprintf(out, "%d.%d.%d.%d", (end_ip & 0xFF000000) >> 0x18, \

            (end_ip & 0x00FF0000) >> 0x10, (end_ip & 0x0000FF00) >> 0x8, end_ip & 0x000000FF);
        getAddress(fp, getValue(fp, i + 4, 3), &country, &location);
        fprintf(out, "\t%s\t%s\n", country, location);
        count++;
    }
    /*ص*/
    return count;
};


/*жһַǷΪַ
ǣ0
ǣ1*/
int beNumber(char c)
{
    if (c >= '0'&&c <= '9')
        return 0;
    else
        return 1;
};


/*Ĳһ洢IPַַ׵ַ
ظIP16ƴ
IPַд󣬺0*/
unsigned long getIP(char *ip_addr)
{
    unsigned long ip = 0;
    int i, j = 0;
    /*ζȡַеĸַ*/
    for (i = 0; i<strlen(ip_addr); i++)
    {
        /*IPַġ.
        ѵǰȡIPֶεֵip
        ע⣬ipΪʱ16Ƶ0x100
        ʱֵ*/
        if (*(ip_addr + i) == '.')
        {
            ip = ip * 0x100 + j;
            j = 0;
        }
        /*ʱд뵱ǰȡIPֶеֵַ
        ӳ10ΪIPַ10*/
        else
        {
            /*жϣIPַ淶10ַ
            0*/
            if (beNumber(*(ip_addr + i)) == 0)
                j = j * 10 + *(ip_addr + i) - '0';
            else
                return 0;
        }
    }
    /*IPֶ4ǡ.ֻ3ӵĸֵֶ*/
    ip = ip * 0x100 + j;
    return ip;
};


void DnsResolverThread(void *p)
{
    TRACE_MSG("DNS resolver thread started.");
    dns_resolver_thread *dnt = (dns_resolver_thread*)p;
    WinMTRNet* wn = dnt->winmtr;

    struct hostent *phent;

    char buf[100];
    int addr = wn->GetAddr(dnt->index);
    sprintf(buf, "%d.%d.%d.%d", (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff, addr & 0xff);

    int haddr = htonl(addr);
    //phent = gethostbyaddr( (const char*)&haddr, sizeof(int), AF_INET);
    FILE *fp;
    unsigned long index_start, index_end, current;
    char *country;
    char *location;
    country = (char*)malloc(MAXBUF);
    location = (char*)malloc(MAXBUF);

    fp = fopen(QQWRY, "rb");
    
    getHead(fp, &index_start, &index_end);
    getAddress(fp, getValue(fp, index_end + 4, 3), &country, &location);
    //IPĿƫ
    current = searchIP(fp, index_start, index_end, addr);
    //ȡIPĹҵַ͵ַ
    getAddress(fp, getValue(fp, current + 4, 3), &country, &location);
    std::string l = std::string(location);
    std::string c = std::string(country);
    std::string h;
    std::string blank = std::string(" ");
    std::string ipaddr = std::string(buf);
    h = ipaddr + blank + c + blank + l;
    char * writable = new char[h.size() + 1];
    std::copy(h.begin(), h.end(), writable);
    writable[h.size()] = '\0';
    wn->SetName(dnt->index, writable);
    delete p;
    fclose(fp);
    TRACE_MSG("DNS resolver thread stopped.");
    _endthread();
}
